
local ON_EVENT = 'OnEvent'
local ON_UPDATE = 'OnUpdate'

local function CompareItems(item1, item2)
  return item1.time > item2.time
end

local function GetScript(self, handler)
  if handler == ON_EVENT then
    return self.__executor.OnEvent
  elseif handler == ON_UPDATE then
    return self.__executor.OnUpdate
  else
    return self.__executor.GetScript(self, handler)
  end
end

local function SetScript(self, handler, func)
  if handler == ON_EVENT then
    self.__executor.OnEvent = func
  elseif handler == ON_UPDATE and self.__executor.size > 0 then
    self.__executor.OnUpdate = func
  else
    self.__executor.SetScript(self, handler, func)
  end
end

local function OnEvent(self, event, ...)
  if self.__executor.OnEvent then
    self.__executor.OnEvent(self, event, ...)
  end
  self.__executor.events[event](self, event, ...)
end

local function OnUpdate(self, ...)
  if self.__executor.OnUpdate then
    self.__executor.OnUpdate(self, ...)
  end
  local time = GetTime()
  while self.__executor.size > 0 do
    local item = self.__executor.queue[self.__executor.size]
    if item.time <= time then
      item.func(unpack(item.args))
      assert(tremove(self.__executor.queue) == item)
      self.__executor.size = self.__executor.size - 1
    else
      break
    end
  end
  if self.__executor.size <= 0 then
    self.__executor.SetScript(self, ON_UPDATE, self.__executor.OnUpdate)
  end
end

local function AddEvent(self, event, func)
  self.__executor.events[event] = func
end

local function RegisterEvents(self)
  for event in pairs(self.__executor.events) do
    self:RegisterEvent(event)
  end
end

local function RunAfter(self, seconds, func, ...)
  local item = {time = GetTime() + seconds, func = func, args = {}}
  for i = 1, select('#', ...) do
    local arg = select(i, ...)
    tinsert(item.args, arg)
  end
  table.bininsert(self.__executor.queue, item, CompareItems)
  self.__executor.size = self.__executor.size + 1
  if self.__executor.size == 1 then
    self.__executor.OnUpdate = self.__executor.GetScript(self, ON_UPDATE)
    self.__executor.SetScript(self, ON_UPDATE, OnUpdate)
  end
end

function SetupExecutor(frame)
  assert(frame)
  assert(frame.__executor == nil)
  assert(frame.AddEvent == nil)
  assert(frame.RegisterEvents == nil)
  assert(frame.RunAfter == nil)
  frame.__executor = {
    events = {},
    queue = {},
    size = 0,
    GetScript = frame.GetScript,
    SetScript = frame.SetScript,
  }
  frame.GetScript = GetScript
  frame.SetScript = SetScript
  frame.AddEvent = AddEvent
  frame.RegisterEvents = RegisterEvents
  frame.RunAfter = RunAfter
  frame.__executor.SetScript(frame, ON_EVENT, OnEvent)
  return frame
end

function CreateExecutor(frameType)
  return SetupExecutor(CreateFrame(frameType or 'Frame'))
end
